/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: mass-lumping.c,v 1.70 2017/01/23 06:46:08 zlb Exp $ */

/*-----------------------------------------------------------------------------
 * This file implements the finite element bases with mass lumping (MLn).
 *
 * Note: the FE bases consist of linear combinations of elementary bases.
 * For elementary bases, we use hierarchical homogeneous monomials as follows:
 * 	For vertex nodes:
 * 		\lambda^O1
 * 	For edge nodes:
 * 		b\cdot\lambda^p, \lambda\in\R^2, p\in\N^2, |p| = O1-2
 * 		(equivalent to \lambda^p, p\in{\N^+}^2, |p| = O1)
 * 	For face nodes:
 * 		b\cdot\lambda^p, \lambda\in\R^3, p\in\N^3, |p| = O2-3
 * 		(equivalent to \lambda^p, p\in{\N^+}^3, |p| = O2)
 * 	For interior nodes:
 * 		b\cdot\lambda^p, \lambda\in\R^4, p\in\N^4, |p| = O3-4
 * 		(equivalent to \lambda^p, p\in{\N^+}^4, |p| = O3)
 *---------------------------------------------------------------------------*/

#include "phg.h"
#include "phg/quad-permu.h"
#include <string.h>		/* memset() */
#include <math.h>		/* pow() */
#include <stdlib.h>		/* qsort() */

enum {S1, S2, S11, S3, S21, S111, S4, S31, S22, S211, S1111};

static const FLOAT *
get_node(DOF_TYPE *type, int n)
/* returns the barycentric coordinates of the 'n'-th node. */
{
    static FLOAT lambda[4];
    int i;

    if (n < NVert * type->np_vert) {		/* Vertex */
	assert(type->np_vert == 1);
	lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.0;
	lambda[n] = type->points[0];
	return lambda;
    }
    n -= NVert * type->np_vert;

    if (n < NEdge * type->np_edge) {		/* Edge */
	i = n / type->np_edge;	/* edge no. */
	n %= type->np_edge;
	lambda[0] = lambda[1] = lambda[2] = lambda[3] = 0.0;
	n = 1 * type->np_vert + 2 * n;
	lambda[GetEdgeVertex(i,0)] = type->points[n];
	lambda[GetEdgeVertex(i,1)] = type->points[n + 1];
	return lambda;
    }
    n -= NEdge * type->np_edge;

    if (n < NFace * type->np_face) {		/* Face */
	i = n / type->np_face;	/* face no. */
	n %= type->np_face;
	lambda[i] = 0.0;
	n = 1 * type->np_vert + 2 * type->np_edge + 3 * n;
	lambda[GetFaceVertex(i,0)] = type->points[n];
	lambda[GetFaceVertex(i,1)] = type->points[n + 1];
	lambda[GetFaceVertex(i,2)] = type->points[n + 2];
	return lambda;
    }
    n -= NFace * type->np_face;

    /* interior */
    assert(n >= 0 && n < type->np_elem);
    return type->points + 1 * type->np_vert
			+ 2 * type->np_edge
			+ 3 * type->np_face
			+ 4 * n;
}

static void
get_monomials(DOF_TYPE *type, const FLOAT *lambda, FLOAT *vals, BOOLEAN grad)
/* evaluates the value or the gradient of all the monomial at barycentric
 * coordinates lambda[] and return them in 'vals' */
{
    int i, n, order, v0, v1, v2;
    BYTE *po;
    FLOAT p0, p1, p2, p3, power[type->order + 1][Dim + 1];

    /* pre-compute and store powers of lambda[] */
    for (n = 0; n < Dim + 1; n++) {
	power[0][n] = 1.0;
	p0 = lambda[n];
	for (i = 0; i < type->order; i++)
	    power[i + 1][n] = power[i][n] * p0;
    }

    order = type->mass_lumping->order0;
    for (i = 0; i < NVert; i++) {		/* Vertices */
	assert(type->np_vert == 1);
	if (!grad) {
	    *vals = power[order][i];
	    vals++;
	}
	else {
	    vals[0] = vals[1] = vals[2] = vals[3] = 0.;
	    if (order >= 1)
		vals[i] = order * power[order - 1][i];
	    vals += 4;
	}
    }

    order = type->mass_lumping->order1;
    for (i = 0; i < NEdge; i++) {		/* Edges */
	for (n = 1; n <= type->np_edge; n++) {
	    if (!grad) {
		*vals = power[n][GetEdgeVertex(i,0)] *
			power[order - n][GetEdgeVertex(i,1)];
		vals++;
	    }
	    else {
		v0 = GetEdgeVertex(i,0);
		v1 = GetEdgeVertex(i,1);
		vals[0] = vals[1] = vals[2] = vals[3] = 0.;
		vals[v0] = n * power[n - 1][v0] *
			   power[order - n][v1];
		vals[v1] = (order - n) * power[n][v0] *
			   power[order - n - 1][v1];
		vals += 4;
	    }
	}
    }

    order = type->mass_lumping->order2;
    for (i = 0; i < NFace; i++) {		/* Faces */
	for (n = 0; n < type->np_face; n++) {
	    po = type->mass_lumping->face_pow + 2 * n;
	    v0 = GetFaceVertex(i,0);
	    v1 = GetFaceVertex(i,1);
	    v2 = GetFaceVertex(i,2);
	    p0 = power[po[0]][v0];
	    p1 = power[po[1]][v1];
	    p2 = power[order - po[0] - po[1]][v2];
	    if (!grad) {
		*vals = p0 * p1 * p2;
		vals++;
	    }
	    else {
		vals[i] = 0.0;
		vals[v0] = po[0] * p1 * p2 * power[po[0] - 1][v0];
		vals[v1] = po[1] * p0 * p2 * power[po[1] - 1][v1];
		vals[v2] = (order - po[0] - po[1]) * p0 * p1 *
			   power[order - po[0] - po[1] - 1][v2];
		vals += 4;
	    }
	}
    }

    /* interior */
    order = type->mass_lumping->order3;
    for (n = 0; n < type->np_elem; n++) {
	po = type->mass_lumping->elem_pow + 3 * n;
	p0 = power[po[0]][0];
	p1 = power[po[1]][1];
	p2 = power[po[2]][2];
	p3 = power[order - po[0] - po[1] - po[2]][3];
	if (!grad) {
	    *vals = p0 * p1 * p2 * p3;
	    vals++;
	}
	else {
	    vals[0] = po[0] * p1 * p2 * p3 * power[po[0] - 1][0];
	    vals[1] = po[1] * p0 * p2 * p3 * power[po[1] - 1][1];
	    vals[2] = po[2] * p0 * p1 * p3 * power[po[2] - 1][2];
	    vals[3] = (order - po[0] - po[1] - po[2]) * p0 * p1 * p2 *
		      power[order - po[0] - po[1] - po[2] - 1][3];
	    vals += 4;
	}
    }
}

static FLOAT *work = NULL;

static int
comp_2float(const void *q1, const void *q2)
/* compares two pairs of FLOATS (for sorting the edge nodes).
 * Note: the nodes are sorted into decreasing order to match DOF_P[1234] */
{
    const FLOAT *p1 = work + 2 * (*(const int *)q1),
		*p2 = work + 2 * (*(const int *)q2);

    return *p1 == *p2 ? 0 : (*p1 > *p2 ? -1 : 1);
}

static void
init(DOF_TYPE *type)
/* this defines the 'Initialize()' member function of DOF_TYPE */
{
    int i;
    FLOAT *pw, *pw0;

#if USE_OMP
    assert(phgThreadId == 0);
#endif	/* USE_OMP */

    if (type->mass_lumping->weights0[0] < 0.0)
	type->mass_lumping->weights0 = NULL;

    /* sort the edge nodes (for phgDofMapEdgeData to work correctly) */
    if (type->np_edge > 1) {
	int ind[type->np_edge];
	FLOAT tmp[2 * type->np_edge];

	for (i = 0; i < type->np_edge; i++)
	    ind[i] = i;
	work = type->points + type->np_vert;
	qsort(ind, type->np_edge, sizeof(ind[0]), comp_2float);
	/* reorder the points */
	memcpy(tmp, work, 2 * type->np_edge * sizeof(*tmp));
	for (i = 0; i < type->np_edge; i++) {
	    work[2 * i + 0] = tmp[2 * ind[i] + 0];
	    work[2 * i + 1] = tmp[2 * ind[i] + 1];
	}
	if (type->mass_lumping->weights0 != NULL) {
	    /* reorder weights0[] accordingly */
	    work = type->mass_lumping->weights0 + type->np_vert;
	    memcpy(tmp, work, type->np_edge * sizeof(*tmp));
	    for (i = 0; i < type->np_edge; i++)
		work[i] = tmp[ind[i]];
	}
	/* reset work */
	work = NULL;
    }

    if (type->mass_lumping->weights0 == NULL)
	return;

    /* set up the weights[] array */
    pw0 = type->mass_lumping->weights0;

    pw = type->mass_lumping->weights = phgAlloc(type->nbas * sizeof(*pw));
    FreeAtExitNoCheck(type->mass_lumping->weights);

    if (type->np_vert > 0) {
	for (i = 0; i < NVert; i++, pw += type->np_vert)
	    memcpy(pw, pw0, type->np_vert * sizeof(*pw));
	pw0 += type->np_vert;
    }

    if (type->np_edge > 0) {
	for (i = 0; i < NEdge; i++, pw += type->np_edge)
	    memcpy(pw, pw0, type->np_edge * sizeof(*pw));
	pw0 += type->np_edge;
    }

    if (type->np_face > 0) {
	for (i = 0; i < NFace; i++, pw += type->np_face)
	    memcpy(pw, pw0, type->np_face * sizeof(*pw));
	pw0 += type->np_face;
    }

    if (type->np_elem > 0)
	memcpy(pw, pw0, type->np_elem * sizeof(*pw));
}

static void
init_buffers(DOF_TYPE *type)
{
    int i, j, k, order;
    BYTE *p;
    FLOAT *A;

    if (type->mass_lumping->face_pow != NULL)
	return;

    /* generate orders of monomials on a face */
    order = type->mass_lumping->order2;
    type->mass_lumping->face_pow = p = phgAlloc(2 * type->np_face * sizeof(*p));
    FreeAtExitNoCheck(type->mass_lumping->face_pow);
    for (i = 1; i < order; i++) {
	for (j = 1; j < order - i; j++) {
	    *(p++) = i;
	    *(p++) = j;
	}
    }
    assert((int)(p - type->mass_lumping->face_pow) == 2 * type->np_face);

    /* generate orders of monomials in the element */
    order = type->mass_lumping->order3;
    type->mass_lumping->elem_pow = p = phgAlloc(3 * type->np_elem * sizeof(*p));
    FreeAtExitNoCheck(type->mass_lumping->elem_pow);
    for (i = 1; i < order; i++) {
	for (j = 1; j < order - i; j++) {
	    for (k = 1; k < order - i - j; k++) {
		*(p++) = i;
		*(p++) = j;
		*(p++) = k;
	    }
	}
    }
    assert((int)(p - type->mass_lumping->elem_pow) == 3 * type->np_elem);

    /* buffers for caching monomials */
    type->mass_lumping->bas_lambda =
	phgCalloc(phgMaxThreads * (Dim + 1), sizeof(FLOAT));
    FreeAtExitNoCheck(type->mass_lumping->bas_lambda);
    type->mass_lumping->bas_values =
	phgAlloc(phgMaxThreads * type->nbas * sizeof(FLOAT));
    FreeAtExitNoCheck(type->mass_lumping->bas_values);

    type->mass_lumping->grad_lambda =
	phgCalloc(phgMaxThreads * (Dim + 1), sizeof(FLOAT));
    FreeAtExitNoCheck(type->mass_lumping->grad_lambda);
    type->mass_lumping->grad_values =
	phgAlloc((Dim + 1) * phgMaxThreads * type->nbas * sizeof(FLOAT));
    FreeAtExitNoCheck(type->mass_lumping->grad_values);

    /* compute the coeficient matrix of the FE bases. */
    type->mass_lumping->coeff =
		phgCalloc(type->nbas * type->nbas, sizeof(FLOAT));
    FreeAtExitNoCheck(type->mass_lumping->coeff);

    A = phgAlloc(type->nbas * type->nbas * sizeof(FLOAT));
    for (i = 0; i < type->nbas; i++) {
	type->mass_lumping->coeff[i + i * type->nbas] = 1.0;
	/* Note: a(i,j) = \psi_j(x_i) */
	get_monomials(type, get_node(type, i), A + i * type->nbas, FALSE);
    }
    phgSolverDenseSolver(type->nbas, type->nbas, A, type->mass_lumping->coeff);
    phgFree(A);
}

static const FLOAT *
ML_bas(DOF *dof, ELEMENT *e, int no0, int no1, const FLOAT *lambda)
{
    static FLOAT **buffers = NULL;
#if USE_OMP
#pragma omp threadprivate(buffers)
#endif  /* USE_OMP */
    int i;
    FLOAT *p, *p0, *bas_lambda, *bas_values, v;
    DOF_TYPE *type = dof->type;
    
    assert(type != NULL && type->mass_lumping != NULL);

    if (buffers == NULL) {
	buffers = phgCalloc(ORDER_MAX, sizeof(*buffers));
	FreeAtExitNoCheck(buffers);
    }

    i = type->mass_lumping->order0;
    assert(i <= ORDER_MAX);
    if (buffers[i] == NULL) {
	buffers[i] = phgCalloc(type->nbas, sizeof(buffers[i][0]));
	FreeAtExitNoCheck(buffers[i]);
    }
    p0 = p = buffers[i];

    if (no0 < 0)
	no0 = 0;
    if (no1 < 0 || no1 > type->nbas)
	no1 = type->nbas;
    if (no1 <= no0)
	return NULL;

#if 0
if (!memcmp(type->name, "TST", 3)) {
    static int *map = NULL;
    DOF_TYPE *type1 = DOF_Pn[type->name[3] - '0'];
    const FLOAT *b;
    if (map == NULL) {
	const FLOAT *pts0 = type->points, *pts1 = type1->points;
	int k = 0, j;
	map = phgAlloc(type->nbas * sizeof(*map));
	for (i = 0; i < NVert; i++, k++)
	    map[k] = k;
	phgPrintf("0 - %g <-> %g\n", (double)pts0[0], (double)pts1[0]);
	pts0 += 1;
	pts1 += 1;
	for (i = 0; i < type->np_edge; i++) {
	    phgPrintf("%d - %g %g <-> %g %g\n", i,
			(double)pts0[2 * i + 0], (double)pts0[2 * i + 1],
			(double)pts1[2 * i + 0], (double)pts1[2 * i + 1]);
	    for (j = 0; j < type->np_edge; j++)
		if (Fabs(pts0[2 * i + 0] - pts1[2 * j + 0]) <= FLOAT_EPSILON &&
		    Fabs(pts0[2 * i + 1] - pts1[2 * j + 1]) <= FLOAT_EPSILON)
		   break;
		assert(j < type->np_edge);
	    map[k + i] = k + j;
	}
	for (i = 1; i < NEdge; i++)
	    for (j = 0; j < type->np_edge; j++)
		map[k + i * type->np_edge + j] = map[k + j] + i * type->np_edge;
	k += NEdge * type->np_edge;
	pts0 += 2 * type->np_edge;
	pts1 += 2 * type->np_edge;
	for (i = 0; i < type->np_face; i++) {
	    phgPrintf("%d - %g %g %g <-> %g %g %g\n", i,
			(double)pts0[3 * i + 0], (double)pts0[3 * i + 1],
			(double)pts0[3 * i + 2],
			(double)pts1[3 * i + 0], (double)pts1[3 * i + 1],
			(double)pts1[3 * i + 2]);
	    for (j = 0; j < type->np_face; j++)
		if (Fabs(pts0[3 * i + 0] - pts1[3 * j + 0]) <= FLOAT_EPSILON &&
		    Fabs(pts0[3 * i + 1] - pts1[3 * j + 1]) <= FLOAT_EPSILON &&
		    Fabs(pts0[3 * i + 2] - pts1[3 * j + 2]) <= FLOAT_EPSILON)
		   break;
		assert(j < type->np_face);
	    map[k + i] = k + j;
	}
	for (i = 1; i < NFace; i++)
	    for (j = 0; j < type->np_face; j++)
		map[k + i * type->np_face + j] = map[k + j] + i * type->np_face;
	k += NFace * type->np_face;
	pts0 += 3 * type->np_face;
	pts1 += 3 * type->np_face;
	for (i = 0; i < type->np_elem; i++) {
	    phgPrintf("%d - %g %g %g %g <-> %g %g %g %g\n", i,
			(double)pts0[4 * i + 0], (double)pts0[4 * i + 1],
			(double)pts0[4 * i + 2], (double)pts0[4 * i + 3],
			(double)pts1[4 * i + 0], (double)pts1[4 * i + 1],
			(double)pts1[4 * i + 2], (double)pts1[4 * i + 3]);
	    for (j = 0; j < type->np_elem; j++)
		if (Fabs(pts0[4 * i + 0] - pts1[4 * j + 0]) <= FLOAT_EPSILON &&
		    Fabs(pts0[4 * i + 1] - pts1[4 * j + 1]) <= FLOAT_EPSILON &&
		    Fabs(pts0[4 * i + 2] - pts1[4 * j + 2]) <= FLOAT_EPSILON &&
		    Fabs(pts0[4 * i + 3] - pts1[4 * j + 3]) <= FLOAT_EPSILON)
		   break;
		assert(j < type->np_elem);
	    map[k + i] = k + j;
	}
	for (i = 0; i < type->nbas; i++)
	    phgPrintf("map[%d] = %d\n", i, map[i]);
    }
    dof->type = type1;
    b = type1->BasFuncs(dof, e, 0, -1, lambda);
    dof->type = type;
    for (i = no0; i < no1; i++)
	*(p++) = b[map[i]];
    return p0;
}
#endif	/* 0 */

    if (type->mass_lumping->coeff == NULL) {
#if USE_OMP
# pragma omp critical (ML_bas)
#endif	/* USE_OMP */
	init_buffers(type);
    }

    bas_lambda = type->mass_lumping->bas_lambda + (Dim + 1) * phgThreadId;
    bas_values = type->mass_lumping->bas_values + type->nbas * phgThreadId;
    if (memcmp(bas_lambda, lambda, (Dim + 1) * sizeof(*lambda))) {
	memcpy(bas_lambda, lambda, (Dim + 1) * sizeof(*lambda));
	get_monomials(type, lambda, bas_values, FALSE);
    }

    for (; no0 < no1; no0++) {
	v = 0.;
	for (i = 0; i < type->nbas; i++)
	    v += type->mass_lumping->coeff[no0  + i * type->nbas]
			* bas_values[i];
	*(p++) = v;
    }

    return p0;
}

static const FLOAT *
ML_grad(DOF *dof, ELEMENT *e, int no0, int no1, const FLOAT *lambda)
{
    static FLOAT **buffers = NULL;
#if USE_OMP
#pragma omp threadprivate(buffers)
#endif  /* USE_OMP */
    int i;
    FLOAT *p, *p0, *grad_lambda, *grad_values, v0, v1, v2, v3;
    DOF_TYPE *type = dof->type;
    
    assert(type != NULL && type->mass_lumping != NULL);

    if (buffers == NULL) {
	buffers = phgCalloc(ORDER_MAX, sizeof(*buffers));
	FreeAtExitNoCheck(buffers);
    }

    i = type->mass_lumping->order0;
    assert(i <= ORDER_MAX);
    if (buffers[i] == NULL) {
	buffers[i] = phgCalloc(type->nbas * (Dim + 1),
					sizeof(buffers[i][0]));
	FreeAtExitNoCheck(buffers[i]);
    }
    p0 = p = buffers[i];

    if (no0 < 0)
	no0 = 0;
    if (no1 < 0 || no1 > type->nbas)
	no1 = type->nbas;
    if (no1 <= no0)
	return NULL;

    if (type->mass_lumping->coeff == NULL) {
#if USE_OMP
# pragma omp critical (ML_grad)
#endif	/* USE_OMP */
	init_buffers(type);
    }

    grad_lambda = type->mass_lumping->grad_lambda + (Dim + 1) * phgThreadId;
    grad_values = type->mass_lumping->grad_values
					+ (Dim + 1) * type->nbas * phgThreadId;
    if (memcmp(grad_lambda, lambda, (Dim + 1) * sizeof(*lambda))) {
	memcpy(grad_lambda, lambda, (Dim + 1) * sizeof(*lambda));
	get_monomials(type, lambda, grad_values, TRUE);
    }

    for (; no0 < no1; no0++) {
	v0 = v1 = v2 = v3 = 0.;
	for (i = 0; i < type->nbas; i++) {
	    v0 += type->mass_lumping->coeff[no0 + i * type->nbas] *
		  grad_values[4 * i];
	    v1 += type->mass_lumping->coeff[no0 + i * type->nbas] *
		  grad_values[4 * i + 1];
	    v2 += type->mass_lumping->coeff[no0 + i * type->nbas] *
		  grad_values[4 * i + 2];
	    v3 += type->mass_lumping->coeff[no0 + i * type->nbas] *
		  grad_values[4 * i + 3];
	}
	*(p++) = v0;
	*(p++) = v1;
	*(p++) = v2;
	*(p++) = v3;
    }

    return p0;
}

/* Note: d = 2 (face) or 3 (element), n = orbit length */
#define MAP_LOC(u0,u1,u2,u3, d, n)	\
	((d == 2 ? u0 * (d + 1) + u1 : (u0 * (d + 1) + u1) * (d + 1) + u2) * n)
#define SET_MAP(u0,u1,u2,u3, d, n, map)					\
    for (i = 0; i < n; i++) {						\
	temp1[0] = temp0[(d + 1) * i + u0];				\
	temp1[1] = temp0[(d + 1) * i + u1];				\
	temp1[2] = temp0[(d + 1) * i + u2];				\
	for (j = 0; j < n; j++)						\
	    if (!memcmp(temp1, temp0 + (d + 1) * j, sizeof(FLOAT) * d))	\
		break;							\
	assert(j < n);							\
	map[MAP_LOC(u0,u1,u2,u3,d,n) + j] = i;			\
    }

#define SET_MAP3(n, map) 					\
    SET_MAP(0,1,2, -1, 2,n,map) SET_MAP(0,2,1, -1, 2,n,map)	\
    SET_MAP(1,0,2, -1, 2,n,map) SET_MAP(1,2,0, -1, 2,n,map)	\
    SET_MAP(2,0,1, -1, 2,n,map) SET_MAP(2,1,0, -1, 2,n,map)

static void
ML_fmap(const DOF_TYPE *type, const ELEMENT *e, int face_no, int *M)
/* maps ordering of local face data to ordering of data in DOF */
{
    static BYTE *map21 = NULL, *map111 = NULL;
    int i, j, k, n, size, v0, v1, v2, u0, u1, u2;

    assert(type != NULL && type->mass_lumping != NULL);

    if (type->np_face <= 1) {
	for (k = 0; k < type->np_face; k++)
	    (*M++) = k;
	return;
    }

    v0 = MapVertex(e, GetFaceVertex(face_no, 0));
    v1 = MapVertex(e, GetFaceVertex(face_no, 1));
    v2 = MapVertex(e, GetFaceVertex(face_no, 2));

    /* map (v0,v1,v2) to (u0,u1,u2), the latter is a permutation of {0,1,2} */
    u0 = u1 = u2 = 0;
    (v0 > v1) ? u0++ : u1++;
    (v0 > v2) ? u0++ : u2++;
    (v1 > v2) ? u1++ : u2++;

    for (n = k = 0; k < type->mass_lumping->norbit; k++) {
	switch (type->mass_lumping->orbits[k]) {
	    case S1:
	    case S2:
	    case S11:
	    case S4:
	    case S31:
	    case S22:
	    case S211:
	    case S1111:
		break;
	    case S3:
		*(M++) = n++;
		break;
	    case S21:
		size = 3;
#if USE_OMP
#pragma omp critical (ML_fmap21)
#endif  /* USE_OMP */
		if (map21 == NULL) {
		    FLOAT temp0[] = {Perm21(.25)};
		    FLOAT temp1[3];
		    map21 = phgCalloc(3 * 3 * size, sizeof(*map21));
		    FreeAtExitNoCheck(map21);
		    SET_MAP3(size, map21)
		}
		j = MAP_LOC(u0, u1, u2, -1, 2, size);
		for (i = 0; i < size; i++)
		    *(M++) = n + map21[j + i];
		n += size;
		break;
	    case S111:
		size = 6;
#if USE_OMP
#pragma omp critical (ML_fmap111)
#endif  /* USE_OMP */
		if (map111 == NULL) {
		    FLOAT temp0[] = {Perm111(.125,.375)};
		    FLOAT temp1[3];
		    map111 = phgCalloc(3 * 3 * size, sizeof(*map111));
		    FreeAtExitNoCheck(map111);
		    SET_MAP3(size, map111)
		}
		j = MAP_LOC(u0, u1, u2, -1, 2, size);
		for (i = 0; i < size; i++)
		    *(M++) = n + map111[j + i];
		n += size;
		break;
	    default:
		phgError(1, "invalid orbit type (%d), abort.\n",
				type->mass_lumping->orbits[k]);
		break;
	}
    }

    assert(n == type->np_face);
}

#define SET_MAP4(n, map)					\
    SET_MAP(0,1,2,3, 3,n,map) SET_MAP(0,2,1,3, 3,n,map)	\
    SET_MAP(1,0,2,3, 3,n,map) SET_MAP(1,2,0,3, 3,n,map)	\
    SET_MAP(2,0,1,3, 3,n,map) SET_MAP(2,1,0,3, 3,n,map)	\
							\
    SET_MAP(0,1,3,2, 3,n,map) SET_MAP(0,2,3,1, 3,n,map)	\
    SET_MAP(1,0,3,2, 3,n,map) SET_MAP(1,2,3,0, 3,n,map)	\
    SET_MAP(2,0,3,1, 3,n,map) SET_MAP(2,1,3,0, 3,n,map)	\
							\
    SET_MAP(0,3,1,2, 3,n,map) SET_MAP(0,3,2,1, 3,n,map)	\
    SET_MAP(1,3,0,2, 3,n,map) SET_MAP(1,3,2,0, 3,n,map)	\
    SET_MAP(2,3,0,1, 3,n,map) SET_MAP(2,3,1,0, 3,n,map)	\
							\
    SET_MAP(3,0,1,2, 3,n,map) SET_MAP(3,0,2,1, 3,n,map)	\
    SET_MAP(3,1,0,2, 3,n,map) SET_MAP(3,1,2,0, 3,n,map)	\
    SET_MAP(3,2,0,1, 3,n,map) SET_MAP(3,2,1,0, 3,n,map)

static void
ML_elem_map(const DOF_TYPE *type, const ELEMENT *e, int *M)
/* maps ordering of local element data to ordering of data in DOF */
{
    static BYTE *map31 = NULL, *map22 = NULL, *map211 = NULL, *map1111 = NULL;
    int i, j, k, n, size, u0, u1, u2;

    assert(type != NULL && type->mass_lumping != NULL);

    if (type->np_elem <= 1 || ElementIsInOrder(e)) {
	for (k = 0; k < type->np_elem; k++)
	    (*M++) = k;
	return;
    }

    u0 = MapVertex(e, 0);
    u1 = MapVertex(e, 1);
    u2 = MapVertex(e, 2);

    for (n = k = 0; k < type->mass_lumping->norbit; k++) {
	switch (type->mass_lumping->orbits[k]) {
	    case S1:
	    case S2:
	    case S11:
	    case S3:
	    case S21:
	    case S111:
		break;
	    case S4:
		*(M++) = n++;
		break;
	    case S31:
		size = 4;
#if USE_OMP
#pragma omp critical (ML_emap31)
#endif  /* USE_OMP */
		if (map31 == NULL) {
		    FLOAT temp0[] = {Perm31(.125)};
		    FLOAT temp1[3];
		    map31 = phgCalloc(4 * 4 * 4 * size, sizeof(*map31));
		    FreeAtExitNoCheck(map31);
		    SET_MAP4(size, map31)
		}
		j = MAP_LOC(u0, u1, u2, 1 - u0 - u1 - u2, 3, size);
		for (i = 0; i < size; i++)
		    *(M++) = n + map31[j + i];
		n += size;
		break;
	    case S22:
		size = 6;
#if USE_OMP
#pragma omp critical (ML_emap22)
#endif  /* USE_OMP */
		if (map22 == NULL) {
		    FLOAT temp0[] = {Perm22(.125)};
		    FLOAT temp1[3];
		    map22 = phgCalloc(4 * 4 * 4 * size, sizeof(*map22));
		    FreeAtExitNoCheck(map22);
		    SET_MAP4(size, map22)
		}
		j = MAP_LOC(u0, u1, u2, 1 - u0 - u1 - u2, 3, size);
		for (i = 0; i < size; i++)
		    *(M++) = n + map22[j + i];
		n += size;
		break;
	    case S211:
		size = 12;
#if USE_OMP
#pragma omp critical (ML_emap211)
#endif  /* USE_OMP */
		if (map211 == NULL) {
		    FLOAT temp0[] = {Perm211(.125,.25)};
		    FLOAT temp1[3];
		    map211 = phgCalloc(4 * 4 * 4 * size, sizeof(*map211));
		    FreeAtExitNoCheck(map211);
		    SET_MAP4(size, map211)
		}
		j = MAP_LOC(u0, u1, u2, 1 - u0 - u1 - u2, 3, size);
		for (i = 0; i < size; i++)
		    *(M++) = n + map211[j + i];
		n += size;
		break;
	    case S1111:
		size = 24;
#if USE_OMP
#pragma omp critical (ML_emap1111)
#endif  /* USE_OMP */
		if (map1111 == NULL) {
		    FLOAT temp0[] = {Perm1111(.0625,.125,.1875)};
		    FLOAT temp1[3];
		    map1111 = phgCalloc(4 * 4 * 4 * size, sizeof(*map1111));
		    FreeAtExitNoCheck(map1111);
		    SET_MAP4(size, map1111)
		}
		j = MAP_LOC(u0, u1, u2, 1 - u0 - u1 - u2, 3, size);
		for (i = 0; i < size; i++)
		    *(M++) = n + map1111[j + i];
		n += size;
		break;
	    default:
		phgError(1, "invalid orbit type (%d), abort.\n",
				type->mass_lumping->orbits[k]);
		break;
	}
    }

    assert(n == type->np_elem);
}

static void
ML_map(const DOF_TYPE *type, const ELEMENT *e, int no, int *M)
{
    if (no < 0)
	ML_elem_map(type, e, M);
    else
	ML_fmap(type, e, no, M);
}

#undef MAP_LOC
#undef SET_MAP
#undef SET_MAP3
#undef SET_MAP4

/*---------------------------------------------------------------------------
 * Rules for transforming the constrained orbits.
 *
 *   Vertex nodes:
 * 	Cons31(0)	-> Cons1(1)		Vertex
 *   Edge nodes:
 * 	Cons22(x)	-> Cons2(0.5)		Edge center
 * 	Cons211(0,b) 	-> Cons11(b)		Edge (s11)
 *   Face nodes:
 *	Cons31(1/3)	-> Cons3(1./3.)		Face center
 *	Cons211(a,0)	-> Cons21(a)		Face (s21)
 *	Cons1111(0,b,c)	-> Cons111(b,c)		Face (s111)
 *
 * The script 'utils/quad2ml' sorts and converts orbits using the rules above.
 *---------------------------------------------------------------------------*/

/* The macro 'DEFINE_ML_TYPE' defines a DOF_TYPE using macros WEIGHTS, POINTS,
 * and ORBITS */

/* The following are C-preprocessor dependent */
#define Orbits(type)	type ## _orbits
#define Weights0(type)	type ## _weights0
#define MLstruct(type)	type ## _MLstruct
#define Points(type)	type ## _points
#define Dof(type)	DOF_ ## type ## _
#define Name(type)	#type

#define	DEFINE_ML_TYPE(O0, O1, O2, O3, type, grad_type)			\
    static int		Orbits(type)[] = {ORBITS};			\
    static FLOAT	Weights0(type)[] = {WEIGHTS};			\
    static FLOAT	Points(type)[] = {POINTS};			\
    static MASS_LUMPING	MLstruct(type) = {				\
	O0, O1, O2, O3, 						\
	sizeof(Orbits(type)) / sizeof(Orbits(type)[0]),			\
	Orbits(type),		/* orbits[] */				\
	Weights0(type),		/* weights0[] */			\
	/* The pointers below are internal buffers */			\
	NULL,			/* weights[] */				\
	NULL, NULL,		/* face_pow, elem_pow */		\
	NULL,			/* coeff */				\
	NULL, NULL,		/* bas_lambda, bas_values */		\
	NULL, NULL		/* grad_lambda, grad_values */		\
    };									\
    DOF_TYPE Dof(type) = {						\
	DofCache, &MLstruct(type), init, Name(type), Points(type),	\
	NULL,						/* orders[] */	\
	grad_type,					/* grad_type */	\
	NULL,						/* base_type */	\
	NULL,						/* hp_info */	\
	phgDofInterpC2FGeneric,				/* Prolong. */	\
	phgDofInterpF2CGeneric,				/* Interp. */	\
	phgDofInitFuncPoint,				/* Ini. func */	\
	ML_bas, ML_grad, ML_map,	 				\
	FE_H1,						/* FE_space */	\
	TRUE,						/* invariant */	\
	FALSE,						/* free_after_use */ \
	-1,						/* id */	\
	4 + 6*(O1-1) + 4*(O2-1)*(O2-2)/2 + (O3-1)*(O3-2)*(O3-3)/6, /* nbas */ \
	O3,						/* order */	\
	0,						/* D_order */	\
	0,						/* continuity */\
	1,						/* dimension */	\
	1, O1-1, (O2-1)*(O2-2)/2, (O3-1)*(O3-2)*(O3-3)/6};

/*-------------------------------- Begin ML1 --------------------------------*/
/* The following macros are generated from the corresponding constrained
 * quadrature rule by the script 'utils/quad2ml' */
#define POINTS	Cons1(1)
#define WEIGHTS	Dup1(.25)
#define ORBITS	S1
DEFINE_ML_TYPE(1, 1, 1, 1, ML1, DOF_DG0)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------- End ML1 ---------------------------------*/

/*-------------------------------- Begin ML2 --------------------------------*/
/* The following macros are generated from the corresponding constrained
 * quadrature rule by the script 'utils/quad2ml' */
#define POINTS \
    Cons1(1),\
    Cons2(0.5),\
    Cons21(.18858048469644503927115437402941689),\
    Perm4(.25)
#define WEIGHTS \
    Dup1(.00129961081762382864323948583189792),\
    Dup2(.00751330903878115632153864252437151),\
    Dup21(.05374664981124274879942069139702288),\
    Dup4(.30476190476190476190476190476190476)
#define ORBITS	S1,S2,S21,S4
DEFINE_ML_TYPE(2, 2, 4, 4, ML2, DOF_DG3)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------- End ML2 ---------------------------------*/

/*-------------------------------- Begin ML3 --------------------------------*/
/* The following macros are generated from the corresponding constrained
 * quadrature rule by the script 'utils/quad2ml' */
#define POINTS \
    Cons1(1),\
    Cons11(.69474012433043399696452214373799579),\
    Cons21(.14804629800083268127155972966379546),\
    Cons21(.42045997555404366358045983652415104),\
    Perm31(.10486452489170353984304348629272939),\
    Perm22(.12587961966825070744487424837645650)
#define WEIGHTS \
    Dup1(.00139318132340935829885454503458246),\
    Dup11(.00439720814497923320460068363824448),\
    Dup21(.00938677154027050153276493215837929),\
    Dup21(.01517875558886845408624234213923300),\
    Dup31(.04276746867938747299136451172239785),\
    Dup22(.07930076278323240149263804629029959)
#define ORBITS	S1,S11,S21,S21,S31,S22
DEFINE_ML_TYPE(3, 3, 5, 6, ML3, DOF_DG5)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------- End ML3 ---------------------------------*/

/*-------------------------------- Begin ML4 --------------------------------*/
/* WARNING: the current rule has negative weights and should not be used */
/* The following macros are generated from the corresponding constrained
 *  * quadrature rule by the script 'utils/quad2ml' */
#define POINTS \
    Cons1(1),\
    Cons11(.80854443961101619213427935540700553),\
    Cons11(.30601867253492915323828632954666129),\
    Cons3(1./3.),\
    Cons21(.25867647251242276280547601945611744),\
    Cons111(.18358911895300261675702374239715381,\
	    .00799833687507618645573675055882157),\
    Perm31(.30062604540622798826003985009648294),\
    Perm31(.12819524754074949404694586327517572),\
    Perm211(.07304272679126745377338651048218843,\
	    .35770884123478518379857662808127490)
#define WEIGHTS \
    Dup1(.00036028843803848727831482031703821),\
/**/Dup11(-.04188920908297572294780990987494447),\
    Dup11(.00437666250190729126960626330898023),\
/**/Dup3(-.02677414236318327749238687899671355),\
    Dup21(.02167000187725150565693278672597593),\
    Dup111(.02421578012002043115377259617036411),\
    Dup31(.06494829681676811150761297509556652),\
    Dup31(.04806667921869404922983168397071333),\
    Dup211(.02187727709367027353926813403705858)
#define ORBITS	S1,S11,S11,S3,S21,S111,S31,S31,S211
DEFINE_ML_TYPE(4, 5, 6, 7, ML4, DOF_DG6)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------- End ML4 ---------------------------------*/

/*-------------------------------- Begin P1B --------------------------------*/
/* This implements the $P_1 \oplus \span{bubble} FE space which may be used
 * to implement, e.g., the multiscale stabilization algorithms:
 * 	Lina Song, Yanren Hou, Haibiao Zheng,
 * 	    A variational multiscale method based on bubble functions
 * 		for convection-dominated convectionCdiffusion equation,
 * 	Applied Mathematics and Computation, 217 (2010) 2226C2237
 *
 * Note: examples/poisson.c does not work when using DG3 as the gradient type.
 */
#define POINTS	Cons1(1), Perm4(.25)
#define WEIGHTS	-1
#define ORBITS	S1,S4
DEFINE_ML_TYPE(1, 1, 1, 4, P1B, DOF_DG0)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------- End P1B ---------------------------------*/

#if TEST_TYPES
/* Some ordinary, non mass-lumping bases for debugging the code */
/*--------------------------------------------------------------------------*/
/* P2-equivalent */
#define POINTS	Cons1(2./2.),		/* vert: 2 */ \
		Cons2(1./2.)		/* edge: 11 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S2
DEFINE_ML_TYPE(2, 2, 2, 2, TST2, DOF_DG1)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* P3-equivalent */
#define POINTS	Cons1 (3./3.),		/* vert: 3 */ \
		Cons11(1./3.),		/* edge: 12 */ \
		Cons3 (1./3.)		/* face: 111 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S3
DEFINE_ML_TYPE(3, 3, 3, 3, TST3, DOF_DG2)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* P4-equivalent */
#define POINTS	Cons1 (4./4.),		/* vert: 4 */ \
		Cons11(1./4.),		/* edge: 13 */ \
		Cons2 (2./4.),		/* edge: 22 */ \
		Cons21(1./4.),		/* face: 112 */ \
		Perm4 (1./4.)		/* elem: 1111 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S2,S21,S4
DEFINE_ML_TYPE(4, 4, 4, 4, TST4, DOF_DG3)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* P5-equivalent */
#define POINTS	Cons1 (5./5.),		/* vert: 5 */ \
		Cons11(1./5.),		/* edge: 14 */ \
		Cons11(2./5.),		/* edge: 23 */ \
		Cons21(1./5.),		/* face: 113 */ \
		Cons21(2./5.),		/* face: 122 */ \
		Perm31(1./5.)		/* elem: 1112 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S11,S21,S21,S31
DEFINE_ML_TYPE(5, 5, 5, 5, TST5, DOF_DG4)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* P6-equivalent */
#define POINTS	Cons1  (6./6.),		/* vert: 6 */ \
		Cons11 (1./6.),		/* edge: 15 */ \
		Cons11 (2./6.),		/* edge: 24 */ \
		Cons2  (3./6.),		/* edge: 33 */ \
		Cons21 (1./6.),		/* face: 114 */ \
		Cons111(1./6.,2./6.),	/* face: 123 */ \
		Cons3  (2./6.),		/* face: 222 */ \
		Perm31 (1./6.),		/* elem: 1113 */ \
		Perm22 (1./6.)		/* elem: 1122 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S11,S2,S21,S111,S3,S31,S22
DEFINE_ML_TYPE(6, 6, 6, 6, TST6, DOF_DG5)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* P7-equivalent */
#define POINTS	Cons1  (7./7.),		/* vert: 7 */ \
		Cons11 (1./7.),		/* edge: 16 */ \
		Cons11 (2./7.),		/* edge: 25 */ \
		Cons11 (3./7.),		/* edge: 34 */ \
		Cons21 (1./7.),		/* face: 115 */ \
		Cons111(1./7.,2./7.),	/* face: 124 */ \
		Cons21 (3./7.),		/* face: 133 */ \
		Cons21 (2./7.),		/* face: 223 */ \
		Perm31 (1./7.),		/* elem: 1114 */ \
		Perm211(1./7.,2./7.),	/* elem: 1123 */ \
		Perm31 (2./7.)		/* elem: 1222 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S11,S11,S21,S111,S21,S21,S31,S211,S31
DEFINE_ML_TYPE(7, 7, 7, 7, TST7, DOF_DG6)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* P8-equivalent */
#define POINTS	Cons1  (8./8.),		/* vert: 8 */ \
		Cons11 (1./8.),		/* edge: 17 */ \
		Cons11 (2./8.),		/* edge: 26 */ \
		Cons11 (3./8.),		/* edge: 35 */ \
		Cons2  (4./8.),		/* edge: 44 */ \
		Cons21 (1./8.),		/* face: 116 */ \
		Cons111(1./8.,2./8.),	/* face: 125 */ \
		Cons111(1./8.,3./8.),	/* face: 134 */ \
		Cons21 (2./8.),		/* face: 224 */ \
		Cons21 (3./8.),		/* face: 233 */ \
		Perm31 (1./8.),		/* elem: 1115 */ \
		Perm211(1./8.,2./8.),	/* elem: 1124 */ \
		Perm22 (1./8.),		/* elem: 1133 */ \
		Perm211(2./8.,1./8.),	/* elem: 1223 */ \
		Perm4  (2./8.)		/* elem: 2222 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S11,S11,S2,S21,S111,S111,S21,S21,S31,S211,S22,S211,S4
DEFINE_ML_TYPE(8, 8, 8, 8, TST8, DOF_DG7)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* P9-equivalent */
#define POINTS	Cons1  (9./9.),		/* vert: 9 */ \
		Cons11 (1./9.),		/* edge: 18 */ \
		Cons11 (2./9.),		/* edge: 27 */ \
		Cons11 (3./9.),		/* edge: 36 */ \
		Cons11 (4./9.),		/* edge: 45 */ \
		Cons21 (1./9.),		/* face: 117 */ \
		Cons111(1./9.,2./9.),	/* face: 126 */ \
		Cons111(1./9.,3./9.),	/* face: 135 */ \
		Cons21 (4./9.),		/* face: 144 */ \
		Cons21 (2./9.),		/* face: 225 */ \
		Cons111(2./9.,3./9.),	/* face: 234 */ \
		Cons3  (3./9.),		/* face: 333 */ \
		Perm31 (1./9.),		/* elem: 1116 */ \
		Perm211(1./9.,2./9.),	/* elem: 1125 */ \
		Perm211(1./9.,3./9.),	/* elem: 1134 */ \
		Perm211(2./9.,4./9.),	/* elem: 1224 */ \
		Perm211(3./9.,1./9.),	/* elem: 1233 */ \
		Perm31 (2./9.)		/* elem: 2223 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S11,S11,S11, S21,S111,S111,S21,S21,S111,S3,\
		S31,S211,S211,S211,S211,S31
DEFINE_ML_TYPE(9, 9, 9, 9, TST9, DOF_DG8)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* P10-equivalent */
#define POINTS	Cons1   (10./10.),		/* vert: A */ \
		Cons11  (1./10.),		/* edge: 19 */ \
		Cons11  (2./10.),		/* edge: 28 */ \
		Cons11  (3./10.),		/* edge: 37 */ \
		Cons11  (4./10.),		/* edge: 46 */ \
		Cons2   (5./10.),		/* edge: 55 */ \
		Cons21  (1./10.),		/* face: 118 */ \
		Cons111 (1./10.,2./10.),	/* face: 127 */ \
		Cons111 (1./10.,3./10.),	/* face: 136 */ \
		Cons111 (1./10.,4./10.),	/* face: 145 */ \
		Cons21  (2./10.),		/* face: 226 */ \
		Cons111 (2./10.,3./10.),	/* face: 235 */ \
		Cons21  (4./10.),		/* face: 244 */ \
		Cons21  (3./10.),		/* face: 334 */ \
		Perm31  (1./10.),		/* elem: 1117 */ \
		Perm211 (1./10.,2./10.),	/* elem: 1126 */ \
		Perm211 (1./10.,3./10.),	/* elem: 1135 */ \
		Perm22  (1./10.),		/* elem: 1144 */ \
		Perm211 (2./10.,1./10.),	/* elem: 1225 */ \
		Perm1111(1./10.,2./10.,3./10.),	/* elem: 1234 */ \
		Perm31  (3./10.),		/* elem: 1333 */ \
		Perm31  (2./10.),		/* elem: 2224 */ \
		Perm22  (2./10.)		/* elem: 2233 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S11,S11,S11,S2,\
		S21,S111,S111,S111,S21,S111,S21,S21,\
		S31,S211,S211,S22,S211,S1111,S31,S31,S22
DEFINE_ML_TYPE(10, 10, 10, 10, TST10, DOF_DG9)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
/* Test basis of orders 2 3 4 5 */
#define POINTS	Cons1  (2./2.),		/* vert: 2 */ \
		Cons11 (1./3.),		/* edge: 12 */ \
		Cons21 (1./4.),		/* face: 112 */ \
		Perm31 (1./5.),		/* elem: 1112 */
#define WEIGHTS	-1.0			/* no weights */
#define ORBITS	S1,S11,S21,S31
DEFINE_ML_TYPE(2, 3, 4, 5, TEST, DOF_DG4)
#undef WEIGHTS
#undef POINTS
#undef ORBITS
/*--------------------------------------------------------------------------*/
#endif	/* TEST_TYPES */

DOF_TYPE *DOF_MLn[] = {DOF_ML1, DOF_ML2, DOF_ML3, DOF_ML4, DOF_P1B, NULL};

#if TEST_TYPES
DOF_TYPE *DOF_TSTn[] =
  {DOF_TST2, DOF_TST3, DOF_TST4, DOF_TST5, DOF_TST6, DOF_TST7, DOF_TST8,
   DOF_TST9, DOF_TST10,DOF_TEST, NULL};
#endif	/* TEST_TYPES */
